from typing import List, Optional, Set, Union

from sqlalchemy.dialects.postgresql import insert

from fittrackee.equipments.models import Equipment, EquipmentType
from fittrackee.users.models import (
    User,
    UserSportPreference,
    UserSportPreferenceEquipment,
)
from fittrackee.utils import decode_short_id
from fittrackee.workouts.models import Sport

from .. import db
from .exceptions import (
    InvalidEquipmentException,
    InvalidEquipmentsException,
    MiscEquipmentLimitExceededException,
)

SPORT_EQUIPMENT_TYPES = {
    "Bike": [
        "Cycling (Sport)",
        "Cycling (Transport)",
        "Cycling (Trekking)",
        "Halfbike",
        "Mountain Biking",
        "Mountain Biking (Electric)",
    ],
    "Bike Trainer": ["Cycling (Virtual)"],
    "Board": ["Standup Paddleboarding", "Windsurfing"],
    "Kayak_Boat": [
        "Canoeing",
        "Canoeing (Whitewater)",
        "Kayaking",
        "Kayaking (Whitewater)",
        "Rowing",
    ],
    "Skates": ["Ice Skating", "Inline Skating"],
    "Paddle": [
        "Canoeing",
        "Canoeing (Whitewater)",
        "Kayaking",
        "Kayaking (Whitewater)",
        "Rowing",
        "Standup Paddleboarding",
    ],
    "Racket": ["Padel (Outdoor)", "Tennis (Outdoor)"],
    "Skis": ["Skiing (Alpine)", "Skiing (Cross Country)"],
    "Shoes": [
        "Hiking",
        "Mountaineering",
        "Padel (Outdoor)",
        "Running",
        "Swimrun",
        "Tennis (Outdoor)",
        "Trail",
        "Walking",
        "Cycling (Sport)",
        "Cycling (Transport)",
        "Cycling (Trekking)",
        "Halfbike",
        "Mountain Biking",
        "Mountain Biking (Electric)",
    ],
    "Snowshoes": ["Snowshoes"],
}
MAX_MISC_LIMIT = 5
TYPE_ERROR_MESSAGE = "equipment_ids must be an array of strings"


def handle_pieces_of_equipment(
    equipment_short_ids: Optional[List[str]],
    auth_user: User,
    sport_id: int,
    existing_equipments: Optional[List[Equipment]] = None,
) -> Union[List[Equipment], None]:
    if equipment_short_ids is None:
        return equipment_short_ids

    equipments_list: List[Equipment] = []
    if not equipment_short_ids:
        return equipments_list

    if not isinstance(equipment_short_ids, list):
        raise InvalidEquipmentsException(TYPE_ERROR_MESSAGE)

    sport = Sport.query.filter_by(id=sport_id).first()
    if not sport:
        raise InvalidEquipmentsException(f"sport id {sport_id} not found")

    equipment_types = []
    misc_count = 0
    for equipment_short_id in equipment_short_ids:
        if not isinstance(equipment_short_id, str):
            raise InvalidEquipmentsException(TYPE_ERROR_MESSAGE)

        equipment = Equipment.query.filter_by(
            uuid=decode_short_id(equipment_short_id), user_id=auth_user.id
        ).first()
        if not equipment:
            raise InvalidEquipmentException(
                status="not_found",
                message=(
                    f"equipment with id {equipment_short_id} does not exist"
                ),
                equipment_short_id=equipment_short_id,
            )

        if not equipment.is_active and (
            not existing_equipments or equipment not in existing_equipments
        ):
            raise InvalidEquipmentException(
                status="inactive",
                message=(
                    f"equipment with id {equipment_short_id} is inactive"
                ),
                equipment_short_id=equipment_short_id,
            )

        if equipment.equipment_type.label == "Misc":
            misc_count += 1
            if misc_count > MAX_MISC_LIMIT:
                raise InvalidEquipmentsException(
                    f"a maximum of {MAX_MISC_LIMIT} pieces of Misc equipment "
                    "can be added"
                )
        else:
            if sport.label not in SPORT_EQUIPMENT_TYPES.get(
                equipment.equipment_type.label, []
            ):
                raise InvalidEquipmentException(
                    status="invalid",
                    message=(
                        f"invalid equipment id {equipment.short_id} "
                        f"for sport {sport.label}"
                    ),
                    equipment_short_id=equipment_short_id,
                )

            if equipment.equipment_type in equipment_types:
                raise InvalidEquipmentsException(
                    "only one piece of equipment per type can be provided"
                )
            equipment_types.append(equipment.equipment_type)

        equipments_list.append(equipment)

    return equipments_list


def handle_default_sports(
    default_for_sport_ids: List[int],
    auth_user: User,
    equipment_type: EquipmentType,
) -> List[UserSportPreference]:
    user_sport_preferences = []
    for sport_id in default_for_sport_ids:
        sport = Sport.query.filter_by(id=sport_id).first()
        if not sport:
            raise InvalidEquipmentsException(
                f"sport (id {sport_id}) does not exist"
            )

        # check if sport is valid for equipment type
        if (
            equipment_type.label != "Misc"
            and sport.label
            not in SPORT_EQUIPMENT_TYPES.get(equipment_type.label, [])
        ):
            raise InvalidEquipmentsException(
                f"invalid sport '{sport.label}' for equipment "
                f"type '{equipment_type.label}'"
            )

        user_sport_preference = UserSportPreference.query.filter_by(
            user_id=auth_user.id,
            sport_id=sport_id,
        ).first()
        if not user_sport_preference:
            user_sport_preference = UserSportPreference(
                user_id=auth_user.id,
                sport_id=sport_id,
                stopped_speed_threshold=sport.stopped_speed_threshold,
                pace_speed_display=sport.pace_speed_display,
            )
            db.session.add(user_sport_preference)
            db.session.flush()
        user_sport_preferences.append(user_sport_preference)
    return user_sport_preferences


def update_user_sport_equipment_preferences_for_misc_if_exist(
    equipment: "Equipment",
    user_sport_preferences: Optional[List["UserSportPreference"]],
    auth_user: "User",
    default_for_sport_ids: list[int],
) -> None:
    if not user_sport_preferences:
        return

    # can not exceeds max limit for Misc type
    sport_ids = []
    for sport_id in default_for_sport_ids:
        query = db.session.query(UserSportPreferenceEquipment).filter(
            UserSportPreferenceEquipment.c.user_id == auth_user.id,
            UserSportPreferenceEquipment.c.sport_id == sport_id,
            UserSportPreferenceEquipment.c.equipment_id != equipment.id,
            (
                UserSportPreferenceEquipment.c.equipment_type_id
                == equipment.equipment_type_id
            ),
        )
        if query.count() >= MAX_MISC_LIMIT:
            sport_ids.append(sport_id)

    if sport_ids:
        raise MiscEquipmentLimitExceededException(
            message=f"a maximum of {MAX_MISC_LIMIT} pieces of Misc "
            f"equipment can be added",
            sport_ids=sport_ids,
        )
    db.session.execute(
        insert(UserSportPreferenceEquipment)
        .values(
            [
                {
                    "equipment_id": equipment.id,
                    "equipment_type_id": equipment.equipment_type_id,
                    "sport_id": sport.sport_id,
                    "user_id": auth_user.id,
                }
                for sport in user_sport_preferences
            ]
        )
        .on_conflict_do_nothing()
    )


def update_user_sport_equipment_preferences_for_non_misc_if_exist(
    equipment: "Equipment",
    user_sport_preferences: Optional[List["UserSportPreference"]],
    auth_user: "User",
    default_for_sport_ids: list[int],
    skip_default_sports_update: Set[int],
) -> None:
    if not user_sport_preferences:
        return

    # remove existing default sports for equipment items with the same
    # equipment type
    db.session.query(UserSportPreferenceEquipment).filter(
        UserSportPreferenceEquipment.c.user_id == auth_user.id,
        UserSportPreferenceEquipment.c.sport_id.in_(
            list(set(default_for_sport_ids) - skip_default_sports_update)
        ),
        (
            UserSportPreferenceEquipment.c.equipment_type_id
            == equipment.equipment_type_id
        ),
    ).delete()

    # create new default sports for equipment item
    values = [
        {
            "equipment_id": equipment.id,
            "equipment_type_id": equipment.equipment_type_id,
            "sport_id": sport.sport_id,
            "user_id": auth_user.id,
        }
        for sport in user_sport_preferences
        if sport.sport_id not in skip_default_sports_update
    ]
    if values:
        db.session.execute(
            insert(UserSportPreferenceEquipment)
            .values(values)
            .on_conflict_do_nothing()
        )


def update_user_sport_equipment_preferences_if_exist(
    equipment: "Equipment",
    user_sport_preferences: Optional[List["UserSportPreference"]],
    auth_user: "User",
    default_for_sport_ids: list[int],
    skip_default_sports_update: Optional[Set[int]] = None,
) -> None:
    if equipment.equipment_type.label == "Misc":
        update_user_sport_equipment_preferences_for_misc_if_exist(
            equipment,
            user_sport_preferences,
            auth_user,
            default_for_sport_ids,
        )
    else:
        if skip_default_sports_update is None:
            skip_default_sports_update = set()
        update_user_sport_equipment_preferences_for_non_misc_if_exist(
            equipment,
            user_sport_preferences,
            auth_user,
            default_for_sport_ids,
            skip_default_sports_update,
        )



"""add user preference to start elevation plots at zero

Revision ID: db58d195c5bf
Revises: 374a670efe23
Create Date: 2023-03-14 04:14:23.781672

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'db58d195c5bf'
down_revision = '374a670efe23'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('start_elevation_at_zero', sa.Boolean(), nullable=True))
    op.execute("UPDATE users SET start_elevation_at_zero = true")
    op.alter_column('users', 'start_elevation_at_zero', nullable=False)


    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('start_elevation_at_zero')

    # ### end Alembic commands ###



import os

import dramatiq
from dramatiq.middleware import Shutdown, TimeLimitExceeded

from fittrackee import db
from fittrackee.constants import TASKS_TIME_LIMIT, TaskPriority
from fittrackee.exceptions import TaskException
from fittrackee.files import get_absolute_file_path
from fittrackee.users.export_data import export_user_data
from fittrackee.users.models import UserTask


def update_task_and_clean(task_id: int) -> None:
    export_request = UserTask.query.filter_by(id=task_id).first()
    if not export_request:
        return

    export_request.errored = True
    export_directory = get_absolute_file_path(
        os.path.join("exports", str(export_request.user_id))
    )
    files = [
        f"user_{name}data.json"
        for name in ["", "workouts_", "equipments_", "comments_"]
    ]
    if export_request.file_path:
        files.append(export_request.file_path)
    for file in files:
        try:
            os.remove(os.path.join(export_directory, file))
        except OSError:
            continue
    db.session.commit()


@dramatiq.actor(
    queue_name="fittrackee_users_exports",
    priority=TaskPriority.MEDIUM,
    time_limit=TASKS_TIME_LIMIT,
    max_retries=0,
    notify_shutdown=True,
)
def export_data(task_id: int) -> None:
    try:
        export_user_data(task_id)
    except (Shutdown, TimeLimitExceeded) as e:
        update_task_and_clean(task_id)
        raise TaskException(
            "upload execution time exceeding limit"
            if isinstance(e, TimeLimitExceeded)
            else "worker shutdown during user data export"
        ) from None
    except Exception as e:
        update_task_and_clean(task_id)
        raise TaskException("error during user data export") from e
    finally:
        db.session.close()



import hashlib
import random
from abc import ABC, abstractmethod
from typing import IO, TYPE_CHECKING, Dict, List, Optional, Union

from flask import current_app
from staticmap3 import Line, StaticMap

from fittrackee import VERSION, appLog, db
from fittrackee.constants import ElevationDataSource
from fittrackee.files import get_absolute_file_path

from ..weather import WeatherService
from .workout_point import WorkoutPoint

if TYPE_CHECKING:
    from datetime import datetime

    from gpxpy.gpx import GPX

    from fittrackee.users.models import User
    from fittrackee.workouts.models import Sport, Workout

weather_service = WeatherService()


class BaseWorkoutWithSegmentsCreationService(ABC):
    # file is converted in gpx format if not in gpx format
    gpx: "GPX"

    @abstractmethod
    def __init__(
        self,
        auth_user: "User",
        workout_file: IO[bytes],
        sport: "Sport",
        # stopped_speed_threshold based on the user's sports preferences,
        # if any. Otherwise, based on the sport
        stopped_speed_threshold: float,
        # in case of refresh (workout is None on creation)
        workout: Optional["Workout"],
        get_weather: bool = True,
        # for refresh
        get_elevation_on_refresh: bool = False,  # for refresh with CLI
        change_elevation_source: Optional[ElevationDataSource] = None,
    ) -> None:
        self.auth_user = auth_user
        self.sport = sport
        self.coordinates: List[List[float]] = []
        self.stopped_speed_threshold = stopped_speed_threshold
        self.workout_name: Optional[str] = None
        self.workout_description: Optional[str] = None
        self.start_point: Optional["WorkoutPoint"] = None
        self.end_point: Optional["WorkoutPoint"] = None
        self.get_weather = get_weather
        self.get_elevation_on_refresh = get_elevation_on_refresh
        self.workout = workout
        self.is_creation = workout is None
        self.change_elevation_source = change_elevation_source

    @abstractmethod
    def get_workout_date(self) -> "datetime":
        pass

    @classmethod
    def get_map_hash(cls, map_filepath: str) -> str:
        """
        Generate a md5 hash used as id instead of workout id, to retrieve map
        image (maps are sensitive data)
        """
        md5 = hashlib.md5(usedforsecurity=False)
        absolute_map_filepath = get_absolute_file_path(map_filepath)
        with open(absolute_map_filepath, "rb") as f:
            for chunk in iter(lambda: f.read(128 * md5.block_size), b""):
                md5.update(chunk)
        return md5.hexdigest()

    @classmethod
    def get_static_map_tile_server_url(cls, tile_server_config: Dict) -> str:
        if tile_server_config["STATICMAP_SUBDOMAINS"]:
            subdomains = tile_server_config["STATICMAP_SUBDOMAINS"].split(",")
            subdomain = f"{random.choice(subdomains)}."  # noqa:S311
        else:
            subdomain = ""
        return tile_server_config["URL"].replace("{s}.", subdomain)

    @classmethod
    def generate_map_image(cls, map_filepath: str, coordinates: List) -> None:
        m = StaticMap(
            width=400,
            height=225,
            padding_x=10,
            headers={"User-Agent": f"FitTrackee v{VERSION}"},
            delay_between_retries=5,
        )
        if not current_app.config["TILE_SERVER"]["DEFAULT_STATICMAP"]:
            m.url_template = cls.get_static_map_tile_server_url(
                current_app.config["TILE_SERVER"]
            )
        line = Line(coords=coordinates, color="#3388FF", width=4)
        m.add_line(line)
        image = m.render()
        image.save(map_filepath)

    @staticmethod
    def get_weather_data(
        start_point: WorkoutPoint, end_point: WorkoutPoint
    ) -> List[Union[Dict, None]]:
        return [
            weather_service.get_weather(start_point),
            weather_service.get_weather(end_point),
        ]

    @abstractmethod
    def _process_file(self) -> "Workout":
        pass

    def process_workout(self) -> "Workout":
        try:
            workout = self._process_file()
        except Exception as e:
            db.session.rollback()
            raise e

        if not self.get_weather:
            return workout

        # Start and end points are None when they have no time
        # When start point has no time, it raises exception before getting
        # weather
        if not self.start_point or not self.end_point:
            appLog.error(
                f"no time for the last point for workout '{workout.short_id}',"
                " no weather data is fetched"
            )
            return workout

        # In case of refresh, it updates only workouts without weather data
        if not workout.weather_start and not workout.weather_end:
            weather_start, weather_end = self.get_weather_data(
                self.start_point,
                self.end_point,
            )
            if weather_start and weather_end:
                workout.weather_start = weather_start
                workout.weather_end = weather_end

        db.session.flush()
        return workout
